int utmp_do_open_session(const char *user, const char *terminal
                        , const char *rhost, pid_t pid
                        , const char **callname, const char **err_descr);
int utmp_do_close_session(const char *terminal
                              , const char **callname, const char **err_descr);
